@@ -277,8 +277,7 @@ class LinuxSCSI(executor.Executor):
         return out.strip()
 
     @staticmethod
-    def is_multipath_running(enforce_multipath,
-                             root_helper,
+    def is_multipath_running(root_helper,
                              execute=None) -> bool:
         try:
             if execute is None:
@@ -289,13 +288,9 @@ class LinuxSCSI(executor.Executor):
             # There was a bug in multipathd where it didn't return an error
             # code and just printed the error message in stdout.
             if out and out.startswith('error receiving packet'):
-                raise putils.ProcessExecutionError('', out, 1, cmd, None)
+                return False
 
-        except putils.ProcessExecutionError as err:
-            if enforce_multipath:
-                LOG.error('multipathd is not running: exit code %(err)s',
-                          {'err': err.exit_code})
-                raise
+        except putils.ProcessExecutionError:
             return False
         return True
 
@@ -384,7 +379,7 @@ class LinuxSCSI(executor.Executor):
             multipath_running = True
         else:
             multipath_running = self.is_multipath_running(
-                enforce_multipath=False, root_helper=self._root_helper)
+                root_helper=self._root_helper)
 
         for device_name in devices_names:
             dev_path = '/dev/' + device_name
@@ -632,8 +627,8 @@ class LinuxSCSI(executor.Executor):
                                     root_helper=self._root_helper)
         return out
 
-    def _multipath_resize_map(self, mpath_id):
-        cmd = ('multipathd', 'resize', 'map', mpath_id)
+    def _multipath_resize_map(self, dm_path):
+        cmd = ('multipathd', 'resize', 'map', dm_path)
         (out, _err) = self._execute(*cmd,
                                     run_as_root=True,
                                     root_helper=self._root_helper)
@@ -644,11 +639,13 @@ class LinuxSCSI(executor.Executor):
 
         return out
 
-    def multipath_resize_map(self, mpath_id):
+    def multipath_resize_map(self, dm_path):
         """Issue a multipath resize map on device.
 
         This forces the multipath daemon to update it's
         size information a particular multipath device.
+
+        :param dm_path: Real path of the DM device (eg: /dev/dm-5)
         """
 
         # "multipathd reconfigure" is async since 0.6.1. While the
@@ -657,7 +654,7 @@ class LinuxSCSI(executor.Executor):
         tstart = time.time()
         while True:
             try:
-                self._multipath_resize_map(mpath_id)
+                self._multipath_resize_map(dm_path)
                 break
             except putils.ProcessExecutionError as err:
                 with excutils.save_and_reraise_exception(reraise=True) as ctx:
@@ -679,6 +676,15 @@ class LinuxSCSI(executor.Executor):
         This function tries to signal the local system's kernel
         that an already attached volume might have been resized.
         """
+        # We need all paths up before extending the devices.
+        # see Launchpad Bug: #2032177 for more details.
+        LOG.debug("Checking paths are valid %s", volume_paths)
+        for volume_path in volume_paths:
+            if not utils.check_valid_device(self, volume_path):
+                LOG.error("Path status is down for path %s", volume_path)
+                raise exception.BrickException("All paths need to be up "
+                                               "to extend the device.")
+
         LOG.debug("extend volume %s", volume_paths)
 
         for volume_path in volume_paths:
@@ -713,7 +719,7 @@ class LinuxSCSI(executor.Executor):
                 LOG.info("mpath(%(device)s) current size %(size)s",
                          {'device': mpath_device, 'size': size})
 
-                self.multipath_resize_map(scsi_wwn)
+                self.multipath_resize_map(os.path.realpath(mpath_device))
 
                 new_size = utils.get_device_size(self, mpath_device)
                 LOG.info("mpath(%(device)s) new size %(size)s",
