@@ -219,6 +219,8 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
                     '-fc-0x1234567890123456-lun-1']
         self.assertEqual(expected, volume_paths)
 
+    @mock.patch.object(
+        base.BaseLinuxConnector, 'check_multipath', mock.MagicMock())
     @mock.patch.object(linuxscsi.LinuxSCSI, 'wait_for_rw')
     @mock.patch.object(os.path, 'exists', return_value=True)
     @mock.patch.object(os.path, 'realpath', return_value='/dev/sdb')
@@ -290,7 +292,10 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
                           self.connector.connect_volume,
                           connection_info['data'])
 
+    @mock.patch.object(
+        base.BaseLinuxConnector, 'check_multipath', mock.MagicMock())
     @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device_path')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'wait_for_mpath_device')
     def _test_connect_volume_multipath(self, get_device_info_mock,
                                        get_scsi_wwn_mock,
                                        get_fc_hbas_info_mock,
@@ -301,6 +306,7 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
                                        find_mp_dev_mock,
                                        access_mode,
                                        should_wait_for_rw,
+                                       wait_mpath_device_mock,
                                        find_mp_device_path_mock):
         self.connector.use_multipath = True
         get_fc_hbas_mock.side_effect = self.fake_get_fc_hbas
@@ -942,3 +948,59 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
                                             True, mock.ANY)
         self.assertEqual(len(expected), realpath_mock.call_count)
         realpath_mock.assert_has_calls(expected)
+
+    @ddt.data(None, mock.sentinel.addressing_mode)
+    @mock.patch.object(fibre_channel.FibreChannelConnector,
+                       '_get_host_devices')
+    @mock.patch.object(fibre_channel.FibreChannelConnector,
+                       '_get_possible_devices')
+    def test__get_possible_volume_paths(self, addressing_mode,
+                                        pos_devs_mock, host_devs_mock):
+        conn_props = {'targets': mock.sentinel.targets}
+        if addressing_mode:
+            conn_props['addressing_mode'] = addressing_mode
+        res = self.connector._get_possible_volume_paths(conn_props,
+                                                        mock.sentinel.hbas)
+        pos_devs_mock.assert_called_once_with(mock.sentinel.hbas,
+                                              mock.sentinel.targets,
+                                              addressing_mode)
+        host_devs_mock.assert_called_once_with(pos_devs_mock.return_value)
+        self.assertEqual(host_devs_mock.return_value, res)
+
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'lun_for_addressing')
+    @mock.patch.object(fibre_channel.FibreChannelConnector, '_get_pci_num')
+    def test__get_possible_devices(self, pci_mock, lun_mock):
+        pci_mock.side_effect = [
+            (mock.sentinel.platform1, mock.sentinel.pci_num1),
+            (mock.sentinel.platform2, mock.sentinel.pci_num2)]
+        hbas = [mock.sentinel.hba1, mock.sentinel.hba2]
+        lun_mock.side_effect = [mock.sentinel.lun1B, mock.sentinel.lun2B] * 2
+        targets = [('wwn1', mock.sentinel.lun1), ('wwn2', mock.sentinel.lun2)]
+
+        res = self.connector._get_possible_devices(
+            hbas, targets, mock.sentinel.addressing_mode)
+
+        self.assertEqual(2, pci_mock.call_count)
+        pci_mock.assert_has_calls([mock.call(mock.sentinel.hba1),
+                                   mock.call(mock.sentinel.hba2)])
+        self.assertEqual(4, lun_mock.call_count)
+        lun_mock.assert_has_calls([
+            mock.call(mock.sentinel.lun1, mock.sentinel.addressing_mode),
+            mock.call(mock.sentinel.lun2, mock.sentinel.addressing_mode)] * 2)
+        expected = [
+            (mock.sentinel.platform1, mock.sentinel.pci_num1, '0xwwn1',
+             mock.sentinel.lun1B),
+            (mock.sentinel.platform1, mock.sentinel.pci_num1, '0xwwn2',
+             mock.sentinel.lun2B),
+            (mock.sentinel.platform2, mock.sentinel.pci_num2, '0xwwn1',
+             mock.sentinel.lun1B),
+            (mock.sentinel.platform2, mock.sentinel.pci_num2, '0xwwn2',
+             mock.sentinel.lun2B),
+        ]
+        self.assertEqual(expected, res)
+
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'is_multipath_running')
+    def test_supports_multipath(self, mock_mpath_running):
+        self.connector.supports_multipath()
+        mock_mpath_running.assert_called_once_with(
+            root_helper=self.connector._root_helper)
